# 
# Copyright (C) 1996-2001 LAAS/CNRS 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    - Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# $LAAS: scout.tcl,v 1.3 2003/02/09 16:31:29 matthieu Exp $
#----------------------------------------------------------------------
package provide Gdhe::models 1.0

#
# Une facette de scout
#
proc scoutFacette {} {
    pushMatrix
    # Positionne la facette verticalement
    translate 0 0 0.21
    rotate 180 0 0 1
    rotate 90 1 0 0 
    # Le rectangle support
    color 20 20 20
    polygon 4 -0.035 0.02 0.035 0.02 0.035 0.135 -0.035 0.135
    # Capteur US
    color 200 200 80
    cylinder 0 0.05 0 z 0.05 0.002
    # support du bumper
    color 20 20 20 
    box 0 0.01 0 0.08 0.02 0.01
    color 150 0 0 
    # Bumper
    box 0 0.01 0.01 0.08 0.02 0.01
    popMatrix
}


proc degToRad {x} {
    return [expr $x*3.1415926535/180.0]
}

objectDelete Scout
objectDelete ScoutCam
objectDelete ScoutCamSupport

proc scout {} {
    concave False
    cullFace True
    object Scout {
	# les 16 capteurs
	for { set i 0 } { $i < 16 } { incr i } {
	    pushMatrix
	    rotate [expr (360.0/16)*$i ] 0 0 1
	    translate 0 0.17 0 
	    scoutFacette
	    popMatrix
	}
	# Calcul des faces superieure et inferieure
	set upper {}
	set lower {}
	for { set i 0 } { $i < 16 } { incr i } {
	    set theta [degToRad [expr (360./16)*$i + 360./32. ]]
	    set x [expr 0.172*cos($theta)]
	    set y [expr 0.172*sin($theta)]
	    set upper [concat $upper $x $y]
	    set lower [concat $x $y $lower]
	}
	# Affichage de faces
	color 20 20 20
	pushMatrix
	translate 0 0 0.215
	eval [concat polygon 16 $lower]
	translate 0 0 0.127
	eval [concat polygon 16 $upper]
	popMatrix
	# La base du robot
	color 20 20 20
	prism 7 \
	    0 0 0.165 \
	    0.14 -0.065 0.055 \
	    0.10 -0.14 0.055 \
	    -0.13 -0.14 0.055 \
	    -0.17 0 0.055 \
	    -0.13 0.14 0.055 \
	    0.10 0.14 0.055 \
	    0.14 0.065 0.055
	# Roue droite 
	color 40 40 40 
	cylinder 0 -0.16 0.10 y -0.20 0.18 0.02 
	cylinder 0 -0.17 0.10 y 0.18 -0.20 0.01
	color 200 200 200
	cylinder 0 -0.171 0.10 y 0.03 0.001
	# Roue gauche
	color 40 40 40
	cylinder 0 0.14 0.10 y 0.18 -0.20 0.02
	cylinder 0 0.16 0.10 y -0.20 0.18 0.01
	color 200 200 200
	cylinder 0 0.17 0.10 y 0.03 0.001
	# Roues folles
	color 40 40 40
	cylinder -0.15 -0.025 0.025 y 0.05 0.02
	cylinder -0.15 0.005 0.025 y 0.05 0.02
	# Antenne
	color 40 40 40
	cylinder 0 0.13 0.34 z 0.005 0.04
	cylinder 0 0.13 0.38 z 0.002 0.015
	sphere 0 0.13 0.395 0.003
    }
}

#
# Camera sur un support Nomadic
#
proc scout_cam {{a1 0} {a2 0}} {
    pushMatrix
    translate 0.11 0 0.34
    object ScoutCamSupport {
	# prtie fixe du support
	color 20 20 20
	cylinder 0 0 0 z -0.03 0.02 0.05
    } 
    translate 0 0 0.05
    rotate $a1 0 0 1
    rotate $a2 0 1 0
    object ScoutCam {
	# PArtie mobile du support
	color 20 20 20
	cylinder 0 0 0 z 0.04 -0.04 0.01
	translate 0.02 0 0.01
	# camera
	color 50 50 50
	box -0.05 0.0 0.0 0.10 0.05 0.04
	color 70 70 70
	cylinder 0.0 0.0 0.022 x 0.028 0.01
	cylinder 0.01 0.0 0.022 x 0.028 0.05 0.02
	cylinder 0.03 0.0 0.022 x 0.05 0.008
	color 200 200 200
	cylinder -0.105 0.01 0.01 x 0.006 0.01
	cylinder -0.105 -0.01 0.01 x 0.006 0.01
    }
    popMatrix
}
