# 
# Copyright (C) 1996-2001 LAAS/CNRS 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    - Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# $LAAS: setup.tcl,v 1.34 2003/06/20 22:15:05 matthieu Exp $
#----------------------------------------------------------------------
package provide Gdhe::setup 1.0

global gdheNoControlPanel

###
### setup - general interface setup
###
proc setup {} {
    global obsDist obsElev obsAzi obsX obsY obsZ
    global xmin ymin zmin xmax ymax zmax
    global polygon_mode perspective_mode
    global gdheBase gdheNoControlPanel

    # gdheBase is the widget in which the main window is
    # displayed. Defaults to ".".
    if {![info exists gdheBase]} { 
	set gdheBase ""
	wm title . "GDHE"
	wm geometry . +0+0
	wm deiconify .
    } elseif { "$gdheBase" != "" } {
	if {![winfo exists $gdheBase]} {
	    error "No such widget: $gdheBase"
	}
    }

    # main window containing the OpenGL widget
    frame $gdheBase.gdhe
    
    set w $gdheBase.gdhe.ogl
    addWindow $w
    togl $w -width 640 -height 480 -privatecmap true -rgba true  \
	-double true -depth true -ident main
    setLights

    # geometry manager for main window
    pack $w  -side left -padx 3 -pady 3 -fill both -expand t
    pack $gdheBase.gdhe  -fill both -expand t

    # biggest possible distance in current world
    set maxdist [expr hypot($ymax - $ymin, $xmax - $xmin)*2.0]

    # Control window
    if {![info exists gdheNoControlPanel]} {
	toplevel $w.pos 
	wm title $w.pos "Position"
	# wm group $w.pos "$gdheBase."
	wm geometry $w.pos -0+0

	# distance between the observer and the scene
	scale  $w.pos.dist  -length 200  -label {Distance} \
	    -from 0 -to $maxdist \
	    -resolution 0.1 \
	    -variable "::obsDist" \
	    -command "setSlider $w dist" -orient horizontal
	# Elevation
	scale  $w.pos.elev   -label {Elevation} -from -180 -to 180 \
	    -variable "::obsElev" \
	    -command "setSlider $w elev" -orient horizontal
	# Azimut
	scale $w.pos.azi -label {Azimut} -from -180 -to 180 \
	    -variable "::obsAzi" \
	    -command "setSlider $w azi" -orient horizontal
	# X Position  
	scale $w.pos.x -label {Target X} -from $xmin -to $xmax \
	    -resolution 0.01\
	    -variable "::obsX" \
	    -command "setSlider $w x" -orient horizontal
	# Y Position
	scale $w.pos.y -label {Target Y} -from $ymin -to $ymax \
	    -resolution 0.01\
	    -variable "::obsY" \
	    -command "setSlider $w y" -orient horizontal
	# Z Position
	scale $w.pos.z -label {Target Z} -from $zmin -to $zmax \
	    -resolution 0.01\
	    -variable "::obsZ" \
	    -command "setSlider $w z" -orient horizontal

	label $w.pos.rate -text rate:

	frame $w.pos.props
	checkbutton $w.pos.props.light -text "Lights" \
	    -variable enable_lighting -command " $w redraw"

	set polygonModeMenu \
	    [tk_optionMenu $w.pos.props.fillStyle polygon_mode fill line]
	$w.pos.props.fillStyle configure -width 10
	$polygonModeMenu entryconfigure 0 -command "setPolygonMode $w"
	$polygonModeMenu entryconfigure 1 -command "setPolygonMode $w"
	set perpectiveModeMenu \
	    [ tk_optionMenu $w.pos.props.perspMode perspective_mode \
		  perspective orthogonal]
	$w.pos.props.perspMode configure -width 10
	$perpectiveModeMenu entryconfigure 0 \
	    -command "$w setObs -perspective ; $w redraw true"
	$perpectiveModeMenu entryconfigure 1 \
	    -command "$w setObs -ortho ; $w redraw true"
    

	grid $w.pos.props.light -column 0 -row 0 -columnspan 2
	grid $w.pos.props.fillStyle -column 0 -row 1
	grid $w.pos.props.perspMode -column 1 -row 1

	# Print button
	button $w.pos.print -text "Dump EPS" -command "print $w"

	# Quit button
	button $w.pos.btn  -text Quit -command exit

	# Geometry management
	pack $w.pos.dist  -fill x
	pack $w.pos.elev  -fill x
	pack $w.pos.azi -fill x
	pack $w.pos.x -fill x
	pack $w.pos.y -fill x
	pack $w.pos.z -fill x
	pack $w.pos.rate -fill x
	pack $w.pos.props -fill x
	pack $w.pos.print -fill x
	pack $w.pos.btn -fill x

    }

    # Handler of mouse events in the main window
    bind $w <ButtonPress-1> {
	set oldX %x
	set oldY %y
    }
    bind $w <ButtonPress-2> {
	set oldX %x
	set oldY %y
    }
    bind $w <ButtonPress-3> {
	set oldX %x
	set oldY %y
    }
    bind $w <B1-Motion> [concat "motion_event_b1 $w" { 
	[lindex [%W config -width] 4] \
	    [lindex [%W config -height] 4] \
	    [expr %x - $oldX] [expr %y - $oldY]
	set oldX %x
	set oldY %y
    }]
    bind $w <B2-Motion> [concat "motion_event_b2 $w $maxdist" {
	[lindex [%W config -width] 4] \
		[lindex [%W config -height] 4] \
		[expr %x - $oldX] [expr %y - $oldY]
	set oldX %x
	set oldY %y
    }]
    bind $w <B3-Motion> [concat "motion_event_b3 $w" {
	[lindex [%W config -width] 4] \
		[lindex [%W config -height] 4] \
		[expr %x - $oldX] [expr %y - $oldY]
	set oldX %x
	set oldY %y
    }]
    # Some keyboard mappings too
    focus $w
    bind $w <Key-Up> "move $w z +"
    bind $w <Key-Down> "move $w z -"
    bind $w <Key-Left> "move $w x -"
    bind $w <Key-Right> "move $w x +"
    bind $w <Shift-Key-Left> "move $w y -"
    bind $w <Shift-Key-Right> "move $w y +"

    # load pre-defined models
    load_env $w;
    
    $w setObs -dist $obsDist -elev $obsElev -azi $obsAzi \
	    -x $obsX -y $obsY -z $obsZ
    $w redraw
}


#----------------------------------------------------------------------

##
## Move the point of attention
##
proc move {w axis sign} {
    upvar \#0 obs[string toupper $axis] pos

    set pos [expr $pos $sign 0.1]
    $w setObs -$axis $pos
    $w redraw
}
	      

#----------------------------------------------------------------------

##
## draw_all - redraw all the scene in a window
##
## called by the Togl widget every time it needs to repaint the window
##
proc draw_all {w} {
    global robots pos platform load trajectory rs_trajectory walls \
	window displayTime transp
    set window $w
    # puts "draw_all $w"
    draw_env 
    draw_per_obst
    # draw the "robots" (in fact any kind of element)
    foreach {r} [array names robots] {
	if {[info exists pos($r)]} {
	    pushMatrix
	    switch [llength $pos($r)] {
		3 {
		    # Three parameters: x y theta
		    set x [lindex $pos($r) 0]
		    set y [lindex $pos($r) 1]
		    set theta [lindex $pos($r) 2]
		    translate $x $y 0
		    rotate $theta 0 0 1
		}
		6 {
		    # Six parameters:  Bryant angles
		    set cap [lindex $pos($r) 0]
		    set roulis [lindex $pos($r) 1]
		    set tangage [lindex $pos($r) 2]
		    set x [lindex $pos($r) 3]
		    set y [lindex $pos($r) 4]
		    set z [lindex $pos($r) 5]
		    translate $x $y $z
		    rotate $cap 0 0 1
		    rotate $roulis 0 1 0
		    rotate $tangage 1 0 0 
		}
		default {
		    puts "r: bizarre llength pos($r) != 3 or 6 [llength $pos($r)]"
		}
	    }
	    eval $robots($r)
	    if {[info exists platform($r)]} {
		eval $platform($r)
	    }
	    if {[info exists load($r)]} {
		eval $load($r)
	    }
	}
	popMatrix

	# Walls attached to the robots (XXX martha specific)
	draw_walls $r

	# A Reeds-Shepp trajectory (XXX)
	if {[info exists rs_trajectory($r)]} {
	    draw_rs_trajectory $r
	}
	if {[info exists trajectory($r)]} {
	    draw_trajectory $r
	}

    }
    # draw translucent parts last.  (XXX not fully implemented yet)
    foreach {r} [array names transp] {
	if {[info exists pos($r)]} {
	    pushMatrix
	    switch [llength $pos($r)] {
		3 {
		    # Three parameters: X Y theta
		    set x [lindex $pos($r) 0]
		    set y [lindex $pos($r) 1]
		    set theta [lindex $pos($r) 2]
		    translate $x $y 0
		    rotate $theta 0 0 1
		}
		6 {
		    # Six parameters: Bryant angles
		    set cap [lindex $pos($r) 0]
		    set roulis [lindex $pos($r) 1]
		    set tangage [lindex $pos($r) 2]
		    set x [lindex $pos($r) 3]
		    set y [lindex $pos($r) 4]
		    set z [lindex $pos($r) 5]
		    translate $x $y $z
		    rotate $cap 0 0 1
		    rotate $roulis 0 1 0
		    rotate $tangage 1 0 0 
		}
		default {
		    puts "bizarre llength pos($r) != 3 or 6 [llength $pos($r)]"
		}
	    }
	    eval $transp($r)
	}
	popMatrix
    }
    # Display the refresh rate if requested
    if {[winfo exists $w.pos.rate]} {
	if {[info exists displayTime]} {
	    if { $displayTime > 10 } {
		set rate [format "%.1f" [expr 1000.0/$displayTime]]
		$w.pos.rate configure -text "rate: $rate frames/s"
	    } else {
		$w.pos.rate configure -text "display time $displayTime ms"
	    }
	} else {
	    $w.pos.rate configure -text ""
	}
    }
}

#----------------------------------------------------------------------

##
## Update the polygon drawing mode of a window and redraw it
##
proc setPolygonMode {w} {
    global polygon_mode

    $w polygonMode $polygon_mode
    $w redraw t
}

#----------------------------------------------------------------------

##
## motion_event - Handler of events in the OpenGL window
##    left button changes azimut and elevation of the observer
##     middle button changes the distance of the observer 
##    right button changes the x and y coordinates of the target point

# left button
proc motion_event_b1 { w width height dx dy } { 
    global obsElev obsAzi

    # compute new  azi and elev
    set elev [expr $obsElev + 180.0 * $dy / $height ]
    if { $elev > 180 } { set elev [expr $elev - 360] }
    if { $elev < -180 } { set elev [expr $elev + 360] }
    set obsElev $elev

    set azi [expr $obsAzi + 360.0 * $dx/$width ]
    if { $azi > 180 } { set azi [expr $azi - 360] }
    if { $azi < -180 } { set azi [expr $azi + 360] }

    # update sliders
    set obsAzi $azi

    # update view 
    $w setObs -elev $obsElev -azi $obsAzi 
    $w redraw
}

# middle button
proc motion_event_b2 { w maxdist width height dx dy } { 
    global obsDist

    # compute new distance
    set dist [expr $obsDist + $maxdist * $dy / $height ]
    if { $dist > $maxdist } { set dist $maxdist }
    if { $dist < 0 } { set dist 0 }

    # update sliders
    set obsDist $dist

    # update view 
    $w setObs -dist $obsDist
    $w redraw
}

# Right button
proc motion_event_b3 { w width height dx dy } { 
    global obsElev obsAzi obsX obsY
    global xmin ymin zmin xmax ymax zmax

    # compute new position
    set azimuth [expr $obsAzi * 3.1415926 / 180]
    set elev [expr $obsElev * 3.1415926 / 180]

    set X [expr $obsX - ($xmax - $xmin) * $dx * cos( $azimuth) / $width ]
    set Y [expr $obsY + ($xmax - $xmin) * $dx * sin( $azimuth) / $width ]

    set X [expr $X  + ($ymax - $ymin) * $dy * sin( $azimuth) / $height ]
    set Y [expr $Y  + ($ymax - $ymin) * $dy * cos( $azimuth) / $height ]


    if { $X > $xmax } { set X $xmax }
    if { $X < $xmin } { set X $xmin }
    if { $Y > $ymax } { set Y $ymax }
    if { $Y < $ymin } { set Y $ymin }

    # update sliders
    set obsX $X
    set obsY $Y

    # update view 
    $w setObs -x $obsX -y $obsY
    $w redraw
}

#----------------------------------------------------------------------

##
## setSlider - Handler of events on sliders in the control panel
##
proc setSlider {w axis value} {

    # Undate variables that changed
    switch -exact $axis {
	dist { $w setObs -dist $value }
	elev { $w setObs -elev $value }
	azi { $w setObs -azi $value }
	x { $w setObs -x $value }
	y { $w setObs -y $value }
	z { $w setObs -z $value }

    }
    # redraw
    $w redraw
}

#----------------------------------------------------------------------
###
### load other files
###
proc load_env {w} {
    
    # objects models
    catch {
	package require Gdhe::models
    } result

    Gdhe::models::init

    # environment data
    package require Gdhe::env

    # lighting data
    package require Gdhe::lights
    Gdhe::lights::setLights
}

#----------------------------------------------------------------------
###
### Definition of GDHE objects
###
### This function uses a global variable called "objects"
### to store the list of defined objects with the associated 
### display list.
###
proc object {name def} {
    global objects

    # Is this object alreadt defined ? 
    if {![info exists objects($name)]} {
	# Allocate a new display list
	set objects($name) [genLists 1]
	# Create the display list
	newList $objects($name)
	# Evaluate the definition of the object in the caller's context
	uplevel 1 $def
	# Close the display list
	endList 
    } else {
	# Already defined object just call the display list.
	callList $objects($name)
    }
}

#----------------------------------------------------------------------
##
## Destruction of an objet
##
proc objectDelete {name} {
    global objects

    if { [info exists objects($name)]} {
	# Free the display list
	deleteLists $objects($name)
	# Free the reference in the object array
	unset objects($name)
    }
}
#----------------------------------------------------------------------
###
### Management of the list of windows
###

#----------------------------------------------------------------------
##
## Add a window to the list
##
proc addWindow {w} {
    global allWindows

    if {[info exists allWindows]} {
	set allWindows [concat $allWindows $w]
    } else {
	set allWindows $w
    }
}

#----------------------------------------------------------------------
##
## Redraw all existing windows
##
## f: true forces an immediate redisplay
##    otherwise wait until main loop is idle
##
proc redrawAllWindows {{f false}} {
    global allWindows

    foreach w $allWindows {
	$w redraw $f
    }    
}

#----------------------------------------------------------------------
##
## Printing dialog
##
proc print {w} {
    set types {
	{{EPS Files} {.eps}}
	{{All Files} *}
    }
    set filename [tk_getSaveFile -title "Save EPS file" -filetypes $types]
    if {$filename != ""} {
	$w dumpEps -color $filename
    }
}

#----------------------------------------------------------------------
###
### This procedure is called by the GDHE server to check if
### a new client can connect. the server passes the IP address
### and/or the hostname of the client, if available.
### 
### This function uses a simple allow/deny scheme. It can be redefined
### to create more sophisticated policies if needed. 
###
proc clientAuthorize {ip name} {
    global allowedClients deniedClients

    foreach c $allowedClients {
	if {[regexp $c $ip] || [regexp $c $name]} {
	    return 
	}
    }
    foreach c $deniedClients {
	if {[regexp $c $ip] || [regexp $c $name]} {
	    error "Connection refused from $ip"
	}
    }
    error "Connection refused from $ip"
}

## Allow access from specicied hosts (regexp)
proc allow {exp} {
    global allowedClients
    
    lappend allowedClients $exp
}

## Deny access from specified hosts (regexp)
proc deny {exp} {
    global deniedClients

    lappend deniedClients $exp
}

#----------------------------------------------------------------------
###
### Startup code
###


# valeurs par defaut
set obsDist 3
set obsElev 0
set obsAzi 0
set obsX 0
set obsY 0
set obsZ 0
# dimensions
set xmin -10.0
set xmax 10.0
set ymin -10.0
set ymax 10.0
set zmin -5.0
set zmax 5.0
  
set polygonMode fill  

#----------------------------------------------------------------------
#  Access control
# 
# Allow rules are evaluated first
#
allow "^127.0.0.1"
allow "^localhost$"

allow "\.laas\.fr$"
allow "^140\.93\."
allow "^195\.83\.132\."
allow "^192\.168\."
deny ".*"
